/* Copyright (c) 1997-2003 Miller Puckette.
* For information on usage and redistribution, and for a DISCLAIMER OF ALL
* WARRANTIES, see the file, "LICENSE.txt," in this distribution.

iemlib2 written by Thomas Musil (c) IEM KUG Graz Austria 2000 - 2003 */

#ifdef NT
#pragma warning( disable : 4244 )
#pragma warning( disable : 4305 )
#endif

#include "m_pd.h"
#include "iemlib.h"
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>

/*
 Beispiel:

 0: 0.1
 1: 0.2
 2: 0.05
 3: 0.3
 4: 0.25
 5: 0.01
 6: 0.15

 index value prev next

 1.) 0.1
 first = 0; ac = 0;
 ?
 first = 0; ac = 1;
 0: 0.1  -1 -1

 2.) 0.2
 first = 0; ac = 1;
 0.2 > 0.1 -> index 0
 first = 1; ac = 2;
 0: 0.1   1 -1
 1: 0.2  -1  0

 3.) 0.05
 first = 1; ac = 2;
 0.05 > 0.2 ----
 0.05 > 0.1 ----
 -> index 2
 first = 1; ac = 3;
 0: 0.1   1  2
 1: 0.2  -1  0
 2: 0.05  0 -1

 4.) 0.3
 first = 1; ac = 3;
 0.3 > 0.2 -> index 0
 first = 3; ac = 4;
 0: 0.1   1  2
 1: 0.2   3  0
 2: 0.05  0 -1
 3: 0.3  -1  1

 5.) 0.25
 first = 3; ac = 4;
 0.25 > 0.3 ----
 0.25 < 0.2 -> index 1
 first = 3; ac = 5;
 0: 0.1   1  2
 1: 0.2   4  0
 2: 0.05  0 -1
 3: 0.3  -1  4
 4: 0.25  3  1

 6.) 0.01
 first = 3; ac = 5;
 0.01 > 0.3 ----
 0.01 > 0.25 ----
 0.01 > 0.2 ----
 0.01 > 0.1 ----
 0.01 > 0.05 ----
 -> index 5
 first = 3; ac = 6;
 0: 0.1   1  2
 1: 0.2   4  0
 2: 0.05  0  5
 3: 0.3  -1  4
 4: 0.25  3  1
 5: 0.01  2 -1

 7.) 0.15
 first = 3; ac = 6;
 0.15 > 0.3 ----
 0.15 > 0.25 ----
 0.15 > 0.2 ----
 0.15 > 0.1 -> index 3
 first = 3; ac = 7;
 0: 0.1   6  2
 1: 0.2   4  6
 2: 0.05  0  5
 3: 0.3  -1  4
 4: 0.25  3  1
 5: 0.01  2 -1
 6: 0.15  1  0


*/



/* ------------------------ list_schedule ---------------------------- */
static t_class *list_schedule_class;

typedef struct _lsched
{
    int        prev;
    int        next;
    int        atom;
    int        leng;
} t_lsched;

typedef struct _list_schedule
{
    t_object   x_obj;
    int        x_size_all;
    int        x_size_sch;
    int        x_ac_all;
    int        x_ac_sch;
    int        x_first;
    int        x_last;
    t_atom     *x_at_all;
    t_lsched   *x_sched;
} t_list_schedule;

static void list_schedule_atcopy(t_atom *src, t_atom *dst, int n)
{
    while(n--)
        *dst++ = *src++;
}

static void list_schedule_list(t_list_schedule *x, t_symbol *s, int ac, t_atom *av)
{
    int i, index, first_index, new_index, prev, next, at_index, it_was_the_smallest_one;
    float cur_val, xval;
    t_lsched *sched=x->x_sched;

    if((ac > 0) && IS_A_FLOAT(av, 0)) /* first item of list has to be a float */
    {
	if((ac+x->x_ac_all) >= x->x_size_all) /* resize buffers */
	{
	    x->x_at_all = (t_atom *)resizebytes(x->x_at_all, x->x_size_all*sizeof(t_atom), 2*(x->x_size_all+ac)*sizeof(t_atom));
	    x->x_size_all = 2*(x->x_size_all+ac);
	}
	if((1+x->x_ac_sch) >= x->x_size_sch)
	{
	    x->x_sched = (t_lsched *)resizebytes(x->x_sched, x->x_size_sch*sizeof(t_lsched), 2*(x->x_size_sch+1)*sizeof(t_lsched));
	    x->x_size_sch = 2*(x->x_size_sch+1);
	}
	list_schedule_atcopy(av, x->x_at_all+x->x_ac_all, ac); /* concaternate atom-list to atom-buffer */
	new_index = x->x_ac_sch;
	sched = x->x_sched;
	sched[new_index].atom = x->x_ac_all;
	sched[new_index].leng = ac;

	at_index = sched[new_index].atom;
	cur_val = atom_getfloat(x->x_at_all + at_index);
	first_index = x->x_first;

	if(new_index == 0)
	{
            sched[0].prev = -1;
	    sched[0].next = -1;
	    x->x_first = 0;
            first_index = x->x_first;
	    x->x_last = 0;
            post("first");
	}
	else if(new_index == 1)
	{
            at_index = sched[first_index].atom;
	    xval = atom_getfloat(x->x_at_all + at_index);
            if(cur_val > xval)
	    {
                sched[first_index].prev = new_index;
		sched[first_index].next = -1;
                sched[new_index].prev = -1;
		sched[new_index].next = first_index;
		x->x_first = new_index;
                first_index = x->x_first;
		x->x_last = first_index;
                post("second was larger");
	    }
	    else
	    {
                sched[first_index].prev = -1;
		sched[first_index].next = new_index;
                sched[new_index].prev = first_index;
		sched[new_index].next = -1;
		x->x_first = first_index;
                first_index = x->x_first;
		x->x_last = new_index;
                post("second was smaller");
	    }
	}
	else
	{
	    index = first_index;
            it_was_the_smallest_one = 1;
	    for(i=0; i<new_index; i++) /* i is only loop-counter */
	    {
		at_index = sched[index].atom;
		xval = atom_getfloat(x->x_at_all + at_index);
		if(cur_val > xval) /* enter new_index at i */
		{
		    next = sched[index].next; /* this is the smaller one */
		    prev = sched[index].prev; /* this is the more greater one */
                    post("i=%d, index=%d, at_index=%d, next=%d, prev=%d, first=%d",i,index,at_index,next,prev,x->x_first);
		    if(prev < 0) /* it is greater than the first one */
		    {
			sched[new_index].prev = -1;
			sched[new_index].next = first_index;
			sched[first_index].prev = new_index;
			x->x_first = new_index;
                        first_index = x->x_first;
		    }
		    else /* between */
		    {
                        sched[new_index].prev = prev;
			sched[new_index].next = index;
			sched[index].prev = new_index;
                        sched[prev].next = new_index;
		    }
		    i = new_index + 1; /* finish this for-loop */
                    it_was_the_smallest_one = 0; /* it was not the smallest one */
		}
		index = sched[index].next; /* the next smallest index */
	    }
	    if(it_was_the_smallest_one)
	    {
                i = x->x_last;
                sched[new_index].prev = i;
		sched[new_index].next = -1;
		sched[i].next = new_index;
                x->x_last = new_index;
	    }
	}
	x->x_ac_all = x->x_ac_all + ac;
	x->x_ac_sch = x->x_ac_sch + 1;
        post("ac_sch=%d, ac_all=%d", x->x_ac_sch, x->x_ac_all);
    }
}

static void list_schedule_float(t_list_schedule *x, t_floatarg f)
{
    t_atom at[2];

    SETFLOAT(at, f);
    list_schedule_list(x, &s_list, 1, at);
}

static void list_schedule_sched_max(t_list_schedule *x)
{
    t_atom *all=x->x_at_all;
    t_lsched *sched=x->x_sched;
    int at_index=x->x_first;
    int ac=x->x_ac_sch;
    int i;

    for(i=0; i<ac; i++)
    {
	outlet_list(x->x_obj.ob_outlet, &s_list, sched[at_index].leng, all+sched[at_index].atom);
	at_index = sched[at_index].next;
	if(at_index == -1)
            i = ac + 1;
    }
}

static void list_schedule_clear(t_list_schedule *x)
{
    x->x_ac_all = 0;
    x->x_ac_sch = 0;
    x->x_first = 0;
    x->x_last = 0;
    x->x_sched[0].prev = -1;
    x->x_sched[0].next = -1;
    x->x_sched[0].atom = 0;
    x->x_sched[0].leng = 0;
}

static void list_schedule_free(t_list_schedule *x)
{
    if(x->x_at_all)
	freebytes(x->x_at_all, x->x_size_all * sizeof(t_atom));
    if(x->x_sched)
        freebytes(x->x_sched, x->x_size_sch * sizeof(t_lsched));
}

static void *list_schedule_new(void)
{
    t_list_schedule *x = (t_list_schedule *)pd_new(list_schedule_class);

    x->x_size_all = 1000;
    x->x_size_sch = 500;
    x->x_at_all = (t_atom *)getbytes(x->x_size_all * sizeof(t_atom));
    x->x_sched = (t_lsched *)getbytes(x->x_size_sch * sizeof(t_lsched));
    list_schedule_clear(x);
    outlet_new(&x->x_obj, &s_list);
    return (x);
}

void list_schedule_setup(void)
{
    list_schedule_class = class_new(gensym("list_schedule"), (t_newmethod)list_schedule_new,
			       (t_method)list_schedule_free, sizeof(t_list_schedule), 0, 0);
    class_addfloat(list_schedule_class, (t_method)list_schedule_float);
    class_addlist(list_schedule_class, list_schedule_list);
    class_addmethod(list_schedule_class, (t_method)list_schedule_sched_max, gensym("sched_max"), 0);
    class_addmethod(list_schedule_class, (t_method)list_schedule_clear, gensym("clear"), 0);
}

